/***************************************************************************
 *   copyright       : (C) 2019-2021 by Pascal Brachet                     *
 *   https://www.xm1math.net/aly/                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   @license GPL-2.0+ <https://spdx.org/licenses/GPL-2.0+.html>           *
 ***************************************************************************/

#ifndef ALY_H
#define ALY_H

#include <QMainWindow>
#include <QDockWidget>
#include <QMenuBar>
#include <QToolBar>
#include <QListWidget>
#include <QStackedWidget>
#include <QTabWidget>
#include <QTreeWidget>
#include <QListWidget>
#include <QLabel>
#include <QAction>
#include <QActionGroup>
#include <QTreeWidgetItem>
#include <QListWidgetItem>
#include <QTableWidgetItem>
#include <QPointer>
#include <QProcess>
#include <QPushButton>
#include <QColor>
#include <QCompleter>
#include <QTextTable>
#include <QVBoxLayout>
#include <QTableWidget>
#include <QTranslator>
#include <QToolButton>
#include <QTimer>
#include <QKeySequence>
#include <QWidgetAction>

#include "alyeditorview.h"
#include "minisplitter.h"
#include "playerbutton.h"
#include "usertagslistwidget.h"
#include "xmltagslistwidget.h"
#include "sourceview.h"
#include "alyconsole.h"
#include "encodingprober/qencodingprober.h"
#include "scandialog.h"
#include "dropshadowlabel.h"


typedef  QMap<AlyEditorView*, QString> FilesMap;
typedef  QMap<QString,QString> KeysMap;
typedef  QString Userlist[10];



class Aly : public QMainWindow
{
    Q_OBJECT

public:
    Aly(QWidget *parent = 0);
    ~Aly();
QString getName();
QFont EditorFont;
QByteArray windowstate,splitter1state,splitter2state,splitter3state,fullscreenstate;
void mainWindowActivated();
//virtual QMenu* createPopupMenu();

public slots:
void load( const QString &f );
void setLine( const QString &line );
void insertFromCommandLine(const QString &entity);
void onOtherInstanceMessage(const QString &);  // For messages for the single instance

private:
void setupMenus();
void setupToolBars();
void createStatusBar();
bool FileAlreadyOpen(QString f);
void closeEvent(QCloseEvent *e);
int LevelItem(const QTreeWidgetItem *item);
int LineItem(const QTreeWidgetItem *item);
bool currentfileSaved();
bool isCurrentModifiedOutside();
bool copyFile(QString origin,QString destination);
//void removeDir(QDir thedir);
void createBuildSubdirectory(QString fn);
QString outputName(QString finame,QString extension);
QString outputBaseName(QString finame);
bool gtkSession();

QString NumberedSource();

int untitled_id;

FilesMap filenames;
KeysMap shortcuts, actionstext;
//gui
QFrame *LeftPanelFrameBis,  *separatorline;

MiniSplitter *splitter1, *splitter2, *splitter3 ;

PlayerButton *toggleStructureButton, *toggleOutputButton, *toggleDiffButton; 
bool winmaximized;

QStackedWidget *EditorView;
QStackedWidget *StackedViewers;


QStackedWidget *LeftPanelStackedWidget;
QListWidget *OpenedFilesListWidget;
UserTagsListWidget *usertagsListWidget;

QTreeWidget *StructureTreeWidget;
QVBoxLayout  *CentralLayoutBis,*LeftPanelLayoutBis;
QHBoxLayout  *LeftPanelLayout, *CentralLayout;

//menu-toolbar
QMenu *fileMenu, *recentMenu, *editMenu, *toolMenu, *sessionMenu;



QMenu *user1Menu, *user11Menu;
QMenu *viewMenu;
QMenu *optionsMenu, *translationMenu, /**appearanceMenu,*/ *settingsMenu, *scriptMenu;
QMenu *helpMenu;
QMenu *defMenu, *ifMenu, *forMenu, *ESMenu;

QToolBar *fileToolBar, *editToolBar, *runToolBar, /* *formatToolBar ,*/ *logToolBar, *LeftPanelToolBar,*LeftPanelToolBarBis, *centralToolBar, *centralToolBarBis;
QAction *recentFileActs[10];
QAction *CleanAct, *StopAct, *UndoAct, *RedoAct, *SaveAct, *CutAct, *CopyAct,*PasteAct, *ToggleDocAct, *ViewStructurePanelAct,  *ViewOutputPanelAct, *ViewDiffPanelAct, *FullScreenAct, *NextDocAct, *PrevDocAct, *ViewOpenedFilesPanelAct ;

QAction *userpanelAct;

QFrame *RightPanelFrame;
QToolBar *RightPanelToolBar;
QVBoxLayout *RightPanelLayout;
DropShadowLabel *titleRightPanel;



QComboBox  *comboFiles;
DropShadowLabel *titleLeftPanel, *posLabel, *stat2, *stat3;

QPushButton *pb1, *pb2, *pb3;


QStringList recentFilesList;
QString sessionTempFile;

//settings
bool eraseSettings, replaceSettings;
QString settingsFileName;
int split1_right, split1_left, split2_top, split2_bottom,  tabwidth;
bool  wordwrap, parenmatch, showline,  showstructview, showoutputview, showdiffview, showfilesview, completion,  modern_style, new_gui,  watchfiles, autosave, tabspaces ;



QString lastDocument, input_encoding, lastChild, lastTemplate, lastScript;
QStringList userCompletionList;
QStringList labelitem, listchildfiles;
Userlist UserMenuName, UserMenuTag;
QString extra_path;
QString python_command;




//dialogs
QPointer<ScanDialog> scanDialog;


SourceView* sourceviewerWidget;
QPointer<AlyConsole> consoleWidget;

//tools
QProcess *proc;
bool FINPROCESS, ERRPROCESS, STOPPROCESS;
QStringList listViewerCommands;


//X11

QString x11style;
QString x11fontfamily;
int x11fontsize;


QStringList userTagsList;

QColor colorBackground, colorLine, colorHighlight, colorNormal, colorComment, colorNumber, colorString, colorOperator, colorIdentifier, colorKeyword;

QList<QColor> edcolors() {QList<QColor> e; e << colorBackground << colorLine << colorHighlight << colorNormal; return e;}
QList<QColor> hicolors() {QList<QColor> h; h << colorNormal << colorComment << colorNumber << colorString << colorOperator << colorIdentifier << colorKeyword; return h;}

QCompleter *completer;
QStringList translationList, scriptList;
QActionGroup *translationGroup/*, *appearanceGroup*/;
QTimer *autosaveTimer;

private slots:
AlyEditorView *currentEditorView() const;
void fileNew();
void fileNewFromFile();
void fileOpen();
void checkModifiedOutsideAll();
void fileSave();
void fileSaveAll();
void fileBackupAll();
void fileSaveAs();
void fileSaveACopy();
void fileClose();
void fileCloseRequested( int index);
void fileCloseAll();
void fileExit();
void fileOpenRecent();
void AddRecentFile(const QString &f);
void UpdateRecentFile();
void CleanRecent();
void filePrint();
void fileOpenAndGoto(const QString &f, int line, bool focus);
void getFocusToEditor();
void fileReload();
void allReload();
void listSelectionActivated(int index);
void ComboFilesInsert(const QString & file);

void ExportToODF();
void ExportToLatex();


void editUndo();
void editRedo();
void editCut();
void editCopy();
void editPaste();
void editSelectAll();
void editFind();
void editFindNext();
void editFindInDirectory();
void editReplace();
void editGotoLine();
void editComment();
void editUncomment();
void editIndent();
void editUnindent();
void editTipTab();
void editRunScript();
void editRunFurnishedScript();



void ReadSettings();
void SaveSettings();
void DeleteSettings();
void CopySettings();
void ReplaceSettings();


void NewDocumentStatus(bool m);
void UpdateCaption();

void UpdateStructure();
void UpdateChildsLabels(QStringList listfiles);
void ParseTree(QTreeWidgetItem *item);
void ItemToRange(QTreeWidgetItem *item);

void ShowOpenedFiles();
void OpenedFileActivated(QListWidgetItem *item);
void ShowStructure();
void ShowUserPanel(); //userPanel
void ClickedOnStructure(QTreeWidgetItem *item,int);

void InsertTag(QString Entity, int dx, int dy);
void InsertUserElement(QListWidgetItem *item);
void InsertFromAction();
void InsertWithSelectionFromAction();
void InsertWithSelectionFromString(const QString& text);
void InsertFromString(const QString& text);





void InsertUserTag(QString Entity);
void InsertUserTag1();
void InsertUserTag2();
void InsertUserTag3();
void InsertUserTag4();
void InsertUserTag5();
void InsertUserTag6();
void InsertUserTag7();
void InsertUserTag8();
void InsertUserTag9();
void InsertUserTag10();
void EditUserMenu();


void ShowDefMenu();
void ShowIfMenu();
void ShowForMenu();
void ShowESMenu();


void RunCommand(QString comd,bool waitendprocess);
void readFromStderr();
void stopProcess();
void readFromStdoutput();
void SlotEndProcess(int err);
void RunPythonScript();
void writeInput(const QString &line);
void CleanOutput();


void OpenTerminal();

void ConvertToUnicode();



/////
void PythonHelp();
void UserManualHelp();
void HelpAbout();
void CheckVersion();


void GeneralOptions();

void gotoNextDocument();
void gotoPrevDocument();

void SetInterfaceFont();

void gotoBookmark1();
void gotoBookmark2();
void gotoBookmark3();


void RemoveUserTag();
void AddUserTag();
void ChangeUserTag();
void UpdateUserTag();

void ModifyShortcuts();

void initCompleter();
void updateCompleter();
void updateTranslation();
// void updateAppearance();

void disableToolsActions();
void enableToolsActions();

void clipboardDataChanged();
void refreshAll();
void refreshAllFromCursor(int newnumlines);
void refreshRange();
void jumpToStructure(int line);


void ToggleStructurePanel();

void ToggleOutputPanel();
void ToggleDiffPanel();
void ToggleFilesPanel();
void ShowStructView(bool change);
void ShowOutputView(bool change);
void ShowDiffView(bool change);
void ShowFilesView(bool change); 
void ToggleFullScreen(); 
void EditUserCompletion();
void showCursorPos(int li, int col);
void customContentsMenuStructure( const QPoint &pos );
void customContentsMenuMain( const QPoint &pos );



void SaveSession();
void LoadSessionFile(const QString &fn);
void LoadSession();
void SaveLastSession();
void LoadLastSession();

void compareDocuments();

protected:
void dragEnterEvent(QDragEnterEvent *event);
void dropEvent(QDropEvent *event);
void keyPressEvent(QKeyEvent * event);
   
};

#endif
