/***************************************************************************
 *   copyright       : (C) 2019-2021 by Pascal Brachet                     *
 *   https://www.xm1math.net/aly/                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   @license GPL-2.0+ <https://spdx.org/licenses/GPL-2.0+.html>           *
 ***************************************************************************/

#ifndef ALYCONSOLE_H
#define ALYCONSOLE_H

#include <QWidget>
#include <QString>
#include <QTextEdit>
#include <QPlainTextEdit>
#include <QTextDocument>
#include <QTextCursor>
#include <QTextBlock>
#include <QProcess>

#include "consolehighlighter.h"





class AlyConsole : public QPlainTextEdit  {
   Q_OBJECT

public:
AlyConsole(QWidget *parent,QFont & efont,QList<QColor> edcolors, QList<QColor> hicolors);
~AlyConsole();


void changeFont(QFont & new_font);
QString getEncoding();
void setEncoding(QString enc);
int getCursorPosition(int parag, int index);
void setCursorPosition(int para, int index);
int numoflines();
int linefromblock(const QTextBlock& p);
void selectword(int line, int col, QString word);
ConsoleHighlighter *highlighter;
void insertNewLine();


QString beginningLine();
int lines() const;
int lineLength(int lineNumber) const;
void setSelection(int startLine, int startCol, int endLine, int endCol);
void removeSelectedText();
void scrollDown();

QString answer;
void output(QString);
//void setTextCursor(const QTextCursor &cursor);

public slots:
void matchAll();

void setColors(QList<QColor> colors);
void insert(const QString &txt, const bool atEnd = false);

private:
QString encoding;


bool matchLeftPar ( QTextBlock currentBlock, int index, int numRightPar );
bool matchRightPar( QTextBlock currentBlock, int index, int numLeftPar );
void createParSelection( int pos );

bool matchLeftBrack ( QTextBlock currentBlock, int index, int numRightBrack );
bool matchRightBrack( QTextBlock currentBlock, int index, int numLeftBrack );
void createBrackSelection( int pos );


QColor colorBackground, colorLine, colorHighlight, colorCursor;

QString prompt;
bool isLocked;
void onEnter();
void printOutput();
    
private slots:
void matchPar();
void matchBrack();
void ensureFinalNewLine();//Qt 4.7.1 bug




protected:
void keyPressEvent(QKeyEvent *event) ;
void paintEvent(QPaintEvent *event);
void contextMenuEvent(QContextMenuEvent *e);
void mouseMoveEvent(QMouseEvent *);




signals:
void sendInput(const QString&);
};

#endif
