/***************************************************************************
 *   copyright       : (C) 2019-2021 by Pascal Brachet                     *
 *   https://www.xm1math.net/aly/                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   @license GPL-2.0+ <https://spdx.org/licenses/GPL-2.0+.html>           *
 ***************************************************************************/

#ifndef ALYEDITORVIEW_H
#define ALYEDITORVIEW_H

#include <qwidget.h>
#include <qfont.h>
#include <qcolor.h>
#include <QStackedWidget>
#include <QSplitter>
#include "alyeditor.h"
#include "linenumberwidget.h"
#include "findwidget.h"
#include "gotolinewidget.h"
#include "replacewidget.h"
#include "minisplitter.h"

class AlyEditorView : public QWidget  {
   Q_OBJECT
public: 
	AlyEditorView(QWidget *parent, QFont & efont,bool line,QList<QColor> edcolors, QList<QColor> hicolors,bool tabspaces=false,int tabwidth=4, QString name="", QStringList ulist=QStringList());
	~AlyEditorView();
  AlyEditor *editor;
  void changeSettings(QFont & new_font,bool line);
  void showFind();
  void showFindNext();
  void showGoto();
  void showReplace();
private slots:
void updateFind();
void updateReplace();
void updateInSelection();
private:
  MiniSplitter *splitter;
  LineNumberWidget* m_lineNumberWidget;
  void setLineNumberWidgetVisible( bool );
  QStackedWidget *Stack;
  FindWidget *findwidget;
  ReplaceWidget *replacewidget;
  GotoLineWidget *gotolinewidget;
};

#endif
