/***************************************************************************
 *   copyright       : (C) 2019-2021 by Pascal Brachet                     *
 *   https://www.xm1math.net/aly/                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   @license GPL-2.0+ <https://spdx.org/licenses/GPL-2.0+.html>           *
 ***************************************************************************/


#include <QtGui>

#include "alyhighlighter.h"
#include "alyeditor.h"
#include "blockdata.h"

AlyHighlighter::AlyHighlighter(QTextDocument *parent)
    : QSyntaxHighlighter(parent)
{
ColorNormal = QColor("#000000");
ColorComment=QColor("#808080");
ColorNumber= QColor("#606060");
ColorString=QColor("#E20800");
ColorOperator=QColor("#8C0000");
ColorIdentifier=QColor("#800080");
ColorKeyword=QColor("#000000");
ColorInOut=ColorNormal;
ColorInOut.setAlpha(150);

m_keywords << "def";
m_keywords << "and";
m_keywords << "del";
m_keywords << "from";
m_keywords << "not";
m_keywords << "while";
m_keywords << "as";
m_keywords << "elif";
m_keywords << "global";
m_keywords << "or";
m_keywords << "with";
m_keywords << "assert";
m_keywords << "else";
m_keywords << "if";
m_keywords << "pass";
m_keywords << "yield";
m_keywords << "break";
m_keywords << "except";
m_keywords << "import";
m_keywords << "print";
m_keywords << "input";
m_keywords << "class";
m_keywords << "exec";
m_keywords << "in";
m_keywords << "raise";
m_keywords << "continue";
m_keywords << "finally";
m_keywords << "is";
m_keywords << "return";
m_keywords << "del";
m_keywords << "for";
m_keywords << "lambda";
m_keywords << "try";
m_keywords << "None";
m_keywords << "True";
m_keywords << "False";



}

AlyHighlighter::~AlyHighlighter(){
}

void AlyHighlighter::SetEditor(AlyEditor *ed)
{
editor=ed;
}


void AlyHighlighter::setColors(QList<QColor> colors)
{
ColorNormal=colors.at(0);
ColorComment=colors.at(1);
ColorNumber=colors.at(2);
ColorString=colors.at(3);
ColorOperator=colors.at(4);
ColorIdentifier=colors.at(5);
ColorKeyword=colors.at(6);
ColorInOut=ColorNormal;
ColorInOut.setAlpha(150);
rehighlight();
}

void AlyHighlighter::highlightBlock(const QString &text)
{
int oldindent=0;
BlockData *previousblockData=static_cast<BlockData *>( currentBlock().previous().userData());
if (previousblockData)
    {
    oldindent=previousblockData->indent;
    }
int newindent;
for (newindent=0; text.at(newindent).isSpace(); newindent++);

BlockData *blockData = new BlockData;



int leftPos = text.indexOf( '(' );
while ( leftPos != -1 ) 
  {
  ParenthesisInfo *info = new ParenthesisInfo;
  info->character = '(';
  info->position = leftPos;

  blockData->insertPar( info );
  leftPos = text.indexOf( '(', leftPos+1 );
  }

int rightPos = text.indexOf(')');
while ( rightPos != -1 ) 
  {
  ParenthesisInfo *info = new ParenthesisInfo;
  info->character = ')';
  info->position = rightPos;

  blockData->insertPar( info );
  rightPos = text.indexOf( ')', rightPos+1 );
  }
leftPos = text.indexOf( '{' );
while ( leftPos != -1 ) 
  {
  ParenthesisInfo *info = new ParenthesisInfo;
  info->character = '{';
  info->position = leftPos;

  blockData->insertBrack( info );
  leftPos = text.indexOf( '{', leftPos+1 );
  }

rightPos = text.indexOf('}');
while ( rightPos != -1 ) 
  {
  ParenthesisInfo *info = new ParenthesisInfo;
  info->character = '}';
  info->position = rightPos;

  blockData->insertBrack( info );
  rightPos = text.indexOf( '}', rightPos+1 );
  }



bool isComment=false;
bool isDefClass=false;
  int i = 0;
  QChar prev, ch;

  const int Standard      = 0;     // Standard text
  const int Digit         = 1;     // Digits
  const int Comment       = 2;     // Comment begins with #
  const int Literal1      = 3;     // String literal beginning with "
  const int Literal2      = 4;     // Other string literal beginning with '
  const int Blockcomment1 = 5;     // Block comments beginning and ending with """
  const int Blockcomment2 = 6;     // Other block comments beginning and ending with '''
  const int ClassName     = 7;     // Text after the keyword class
  const int DefineName    = 8;     // Text after the keyword def

  int endStateOfLastPara = previousBlockState();
  if (endStateOfLastPara < 0 || endStateOfLastPara > 8) 
    endStateOfLastPara = Standard;

  while ( i < text.length() )
  {
    ch = text.at( i );

    switch ( endStateOfLastPara )
    {
    case Standard:
      {
        switch ( ch.unicode() )
        {
        case '#':
          {
            // begin a comment
            setFormat( i, 1, ColorComment);
            endStateOfLastPara=Comment;
          } break;
        case '"':
          {
            // Begin either string literal or block comment
            if ((i>=2) && text.at(i-1) == QLatin1Char('"') &&
                text.at(i-2) == QLatin1Char('"'))
            {
              setFormat( i-2, 3, ColorComment);
              endStateOfLastPara=Blockcomment1;
            }
            else
            {
              setFormat( i, 1, ColorString);
              endStateOfLastPara=Literal1;
            }
          } break;
        case '\'':
          {
            // Begin either string literal or block comment
            if ((i>=2) && text.at(i-1) == QLatin1Char('\'') && 
                text.at(i-2) == QLatin1Char('\''))
            {
              setFormat( i-2, 3, ColorComment);
              endStateOfLastPara=Blockcomment2;
            }
            else
            {
              setFormat( i, 1, ColorString);
              endStateOfLastPara=Literal2;
            }
          } break;
        case ' ':
          {
            // ignore whitespaces
          } break;
        case '\t':
          {
            QTextCharFormat tabFormat;
                tabFormat.setUnderlineColor(ColorInOut);
                tabFormat.setUnderlineStyle (QTextCharFormat::DotLine);
                setFormat( i,1, tabFormat);
          } break;
        case '(': case ')': case '[': case ']': 
        case '+': case '-': case '*': case '/': 
        case ':': case '%': case '^': case '~': 
        case '!': case '=': case '<': case '>': // possibly two characters
          {
            setFormat(i, 1, ColorOperator);
            endStateOfLastPara=Standard;
          } break;
        default:
          {
            // Check for normal text
            if ( ch.isLetter() || ch == QLatin1Char('_') )
            {
              QString buffer;
              int j=i;
              while ( ch.isLetterOrNumber() || ch == QLatin1Char('_') ) {
                buffer += ch;
                ++j;
                if (j >= text.length())
                  break; // end of text
                ch = text.at(j);
              }

              if ( m_keywords.contains( buffer ) != 0 ) {
                if ( buffer == QLatin1String("def")) {
                  endStateOfLastPara = DefineName;isDefClass=true;}
                else if ( buffer == QLatin1String("class")) {
                  endStateOfLastPara = ClassName;isDefClass=true;}

                QTextCharFormat keywordFormat;
                if (buffer== QLatin1String("print") || buffer== QLatin1String("input")) {
                    keywordFormat.setForeground(ColorInOut);
                }
                else keywordFormat.setForeground(ColorKeyword);
                keywordFormat.setFontWeight(QFont::Bold);
                setFormat( i, buffer.length(), keywordFormat);
              }
              else {
                setFormat( i, buffer.length(),ColorNormal);
              }

              // increment i
              if ( !buffer.isEmpty() )
                i = j-1;
            }
            // this is the beginning of a number
            else if ( ch.isDigit() )
            {
              setFormat(i, 1, ColorNumber);
              endStateOfLastPara=Digit;
            }
            // probably an operator
            else if ( ch.isSymbol() || ch.isPunct() )
            {
              setFormat( i, 1, ColorOperator);
            }
          }
        }
      } break;
    case Comment:
      {
       isComment=true;   
        setFormat( i, 1, ColorComment);
      } break;
    case Literal1:
      {
        setFormat( i, 1, ColorString);
        if ( ch == QLatin1Char('"') )
          endStateOfLastPara = Standard;
      } break;
    case Literal2:
      {
        setFormat( i, 1, ColorString);
        if ( ch == QLatin1Char('\'') )
          endStateOfLastPara = Standard;
      } break;
    case Blockcomment1:
      {
        isComment=true; 
        setFormat( i, 1, ColorComment);
        if ( i>=2 && ch == QLatin1Char('"') &&
            text.at(i-1) == QLatin1Char('"') &&
            text.at(i-2) == QLatin1Char('"'))
          endStateOfLastPara = Standard;
      } break;
    case Blockcomment2:
      {
        isComment=true;   
        setFormat( i, 1, ColorComment);
        if ( i>=2 && ch == QLatin1Char('\'') &&
            text.at(i-1) == QLatin1Char('\'') &&
            text.at(i-2) == QLatin1Char('\''))
          endStateOfLastPara = Standard;
      } break;
    case DefineName:
      {
        if ( ch.isLetterOrNumber() || ch == QLatin1Char(' ') || ch == QLatin1Char('_') )
        {
          setFormat( i, 1, ColorIdentifier);
        }
        else
        {
          if ( ch.isSymbol() || ch.isPunct() )
            setFormat(i, 1, ColorOperator);
          endStateOfLastPara = Standard;
        }
      } break;
    case ClassName:
      {
        if ( ch.isLetterOrNumber() || ch == QLatin1Char(' ') || ch == QLatin1Char('_') )
        {
          setFormat( i, 1, ColorIdentifier);
        }
        else
        {
          if (ch.isSymbol() || ch.isPunct() )
            setFormat( i, 1, ColorOperator);
          endStateOfLastPara = Standard;
        }
      } break;
    case Digit:
      {
        if (ch.isDigit() || ch == QLatin1Char('.'))
        {
          setFormat( i, 1, ColorNumber);
        }
        else
        {
          if ( ch.isSymbol() || ch.isPunct() )
            setFormat( i, 1, ColorOperator);
          endStateOfLastPara = Standard;
        }
      }break;
    }

    prev = ch;
    i++;
  }
//****************************************************

if (text.trimmed().length()>0) blockData->indent=newindent;
else blockData->indent=oldindent;
setCurrentBlockUserData(blockData);



if (blockData->indent>oldindent )
  {
  BlockInfo *info = new BlockInfo;
  info->character = 'b';
  info->position = 0;
  info->indent=blockData->indent;
  info->open=true;
  if (previousblockData) previousblockData->insertLat( info );
  }
//else
if (blockData->indent<oldindent)
  {
  BlockInfo *info = new BlockInfo;
  info->character = 'e';
  info->position = text.length()-1;
  info->indent=oldindent;
  info->open=false;
    for (QTextBlock block = currentBlock().previous(); block.isValid(); block = block.previous())   
        {
        BlockData *data=static_cast<BlockData *>( block.userData());    
        if (data)
            {
            QVector<BlockInfo *> infos = data->blocks();
            for ( int it=0; it<infos.size(); ++it ) 
                {
                BlockInfo *inf=infos.at(it);
                if (inf->character=='b' &&  inf->indent>blockData->indent && inf->open && previousblockData) 
                    {
                    previousblockData->insertLat( info );
                    inf->open=false;
                    }
                }
            }
        }
      
  }
if (currentBlock().blockNumber()==document()->blockCount()-1)
  {
  BlockInfo *info = new BlockInfo;
  info->character = 'e';
  info->position = text.length()-1;
  info->indent=oldindent;
  info->open=false;
    for (QTextBlock block = currentBlock().previous(); block.isValid(); block = block.previous())   
        {
        BlockData *data=static_cast<BlockData *>( block.userData());    
        if (data)
            {
            QVector<BlockInfo *> infos = data->blocks();
            for ( int it=0; it<infos.size(); ++it ) 
                {
                BlockInfo *inf=infos.at(it);
                if (inf->character=='b' && inf->open) 
                    {
                    blockData->insertLat( info );
                    inf->open=false;
                    }
                }
            }
        }
      
  }
 //**************************************************** 
  // only block comments can have several lines
  if ( endStateOfLastPara != Blockcomment1 && endStateOfLastPara != Blockcomment2 ) 
  {
    endStateOfLastPara = Standard ;
  } 

  setCurrentBlockState(endStateOfLastPara);

  i=currentBlock().blockNumber();
editor->checkStructUpdate(document(),currentBlock().position(),text,i);
// const QList<StructItem>& structure = editor->getStructItems();
// for (int j = 0; j < structure.count(); j++)
// {
// if ((structure.at(j).cursor.block().blockNumber()==currentBlock().blockNumber()) && (structure.at(j).type!=0)) 
//   {
//   int offwidth=0;
//   QTextCursor curs = editor->document()->find("{",structure.at(j).cursor.position());
//   if (!curs.isNull()  && (curs.position()<currentBlock().length()+currentBlock().position()))
//     {
//     offwidth=curs.position()-structure.at(j).cursor.position()+structure.at(j).item.length()+1;
//     }
//     if (offwidth>0) setFormat(structure.at(j).cursor.position()-currentBlock().position(),offwidth,structFormat);
//   }
  
}

bool AlyHighlighter::isSpace(QChar c) const
{
    return c == QLatin1Char(' ')
        || c == QChar::Nbsp
        || c == QChar::LineSeparator
        || c == QLatin1Char('\t')
        ;
}


